package org.supermind.crawl;

import org.apache.nutch.util.LogFormatter;

import java.net.URL;
import java.util.Comparator;
import java.util.HashMap;
import java.util.TreeSet;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Fetchlist that assigns higher priority to hosts with many pending pages.
 */
public class DefaultFetchList implements FetchList {
  protected static Logger LOG =
      LogFormatter.getLogger(FetchList.class.getName());

  /**
   * Total number of URLs in this fetchlist.
   */
  private int currentSize;

  /**
   * {@link LastModifiedDB}.
   */
  protected LastModifiedDB lastModifiedDb = LastModifiedDB.NO_OP;

  /**
   * Interval between consecutive accesses to same host =
   * time taken to download request * waitFactor.
   */
  private int waitFactor = 0;

  /**
   * A customized data structure should really be used here, to reduce the memory
   * overhead of maintaining a hashmap and a treeset.
   */
  private HashMap<String, HostQueue> availableHosts = new HashMap<String, HostQueue>();

  /**
   * Available host names, sorted by number of pages in each HostQueue.
   * When the number of pages change, remove and re-add the hostname for a
   * re-sort.
   */
  private TreeSet<String> availableHostNames = new TreeSet<String>(new Comparator<String>() {
    public int compare(String s, String s1) {
      if (s.equals(s1)) return 0;
      HostQueue h = availableHosts.get(s);
      HostQueue h1 = availableHosts.get(s1);
      if (h == null || h1 == null) return -1;

      int hSize = h.getPages().size();
      int hSize1 = h1.getPages().size();
      if (hSize > hSize1) return -1;
      if (hSize < hSize1) return 1;

      return 1;
    }
  });

  /**
   * Blocked HostQueues.
   */
  private HashMap<String, HostQueue> waitedHosts = new HashMap();

  /**
   * Contains all entries in waitedHosts which have nextAvailable > -1, ordered by increasing time.
   */
  private TreeSet<HostQueue> waitedHostQueue = new TreeSet(new Comparator<HostQueue>() {
    public int compare(HostQueue hostQueue, HostQueue hostQueue1) {
      if (hostQueue.equals(hostQueue1)) return 0;
      if (hostQueue.getNextAvailable() > hostQueue1.getNextAvailable()) return 1;
      if (hostQueue.getNextAvailable() < hostQueue1.getNextAvailable()) return -1;
      return 1;
    }
  });

  private void cleanExpiredServerBlocks() {
    while (!waitedHostQueue.isEmpty()) {
      HostQueue hostQueue = waitedHostQueue.first();
      long time = hostQueue.getNextAvailable();
      if (time > -1 &&                          // if -1, host not released yet
          System.currentTimeMillis() >= time) {
        avail(hostQueue);
        if (LOG.isLoggable(Level.FINER)) LOG.finer("Re-adding " + hostQueue.getName());
      } else {
        break;
      }
    }
  }

  /**
   * Move a hostqueue from the wait queue to the available queue.
   *
   * @param hostQueue
   */
  private void avail(HostQueue hostQueue) {
    hostQueue.setNextAvailable(-1);
    final String hostname = hostQueue.getName();
    waitedHosts.remove(hostname);
    waitedHostQueue.remove(hostQueue);

    // the order of calls here is important because availableHostNames
    // assumes a corresponding entry in availableHOsts
    availableHosts.put(hostname, hostQueue);
    availableHostNames.add(hostname);
  }


  public boolean contains(URL url) {
    String hostname = url.getHost();
    HostQueue hostQueue = availableHosts.get(hostname);
    if (hostQueue == null && waitedHosts.size() > 0)
      hostQueue = waitedHosts.get(hostname);

    return hostQueue == null ? false : hostQueue.contains(url);
  }

  public HostQueue next() {
    cleanExpiredServerBlocks();
    if (availableHosts.size() == 0) {
//      if (LOG.isLoggable(Level.FINEST)) LOG.finest("No more hosts available. WaitedHosts:" + waitedHostQueue);
      if (waitedHostQueue.size() > 0 || waitedHosts.size() > 0) {
        long start = System.currentTimeMillis();
        while (availableHosts.size() == 0)
          try {
            synchronized (this) {
              wait(500);
              cleanExpiredServerBlocks();
            }
          } catch (InterruptedException e) {
          }
        long stop = System.currentTimeMillis();
        if (LOG.isLoggable(Level.FINE))
          LOG.fine("Waited " + (stop - start) + "ms for available host.");
      } else {
        return null;
      }
    }

    HostQueue hostQueue = availableHosts.get(availableHostNames.first());
    if (LOG.isLoggable(Level.FINEST))
      LOG.finest("Pages in " + hostQueue.getName() + " : " + hostQueue.getPages().size());

    // if hostQueue with the most pages has no more pages,
    // then this fetchlist is empty
    if (hostQueue.getPages().size() == 0) {
      return null;
    }
    unavail(hostQueue);
    return hostQueue;
  }

  /**
   * Remove a hostqueue from the list of unused/available hosts.
   *
   * @param hostQueue
   */
  private void unavail(HostQueue hostQueue) {
    availableHosts.remove(hostQueue.getName());
    availableHostNames.remove(hostQueue.getName());
    waitedHosts.put(hostQueue.getName(), hostQueue);
    hostQueue.setNextAvailable(-1);
  }

  public void release(HostQueue hostQueue, int popped, long timeTaken) {
    if (LOG.isLoggable(Level.FINER)) LOG.finer("Releasing " + hostQueue.getName());
    hostQueue.setNextAvailable(getNextAvailable(timeTaken / popped));
    waitedHostQueue.add(hostQueue);
    currentSize -= popped;
  }

  /**
   * When a HostQueue is next available.
   *
   * @param timeTaken time taken to download a page
   * @return date (in milliseconds)
   */
  protected long getNextAvailable(long timeTaken) {
    return System.currentTimeMillis() + (timeTaken * waitFactor);
  }

  public void queue(ScheduledURL parent, URL url) {
    if (LOG.isLoggable(Level.FINER))
      LOG.finer("Queuing " + url + " from " + parent.url);

    ScheduledURL scheduledURL = new ScheduledURL(parent.seedIndex, parent.id, url);
    // TODO: lastmodified checking should be batched
    scheduledURL.lastModified = lastModifiedDb.getLastModified(url);
    String hostname = url.getHost();
    HostQueue hostQueue = availableHosts.get(hostname);
    final boolean available = hostQueue != null;
    if (!available && waitedHosts.size() > 0) {
      hostQueue = waitedHosts.get(hostname);
    }

    if (hostQueue == null) {
      if (LOG.isLoggable(Level.FINEST))
        LOG.finest("Creating new HostQueue for " + hostname);
      hostQueue = new HostQueue(hostname, url.getProtocol());
      hostQueue.add(scheduledURL);
      availableHosts.put(hostname, hostQueue);
      availableHostNames.add(hostname);
    } else {
      if (available) {
        availableHostNames.remove(hostname); // must be removed before adding url for treeset's benefit
        hostQueue.add(scheduledURL);
        availableHostNames.add(hostname);
      } else {
        hostQueue.add(scheduledURL);
      }
    }
    ++currentSize;
  }

  public void init() {

  }

  public void close() {
    availableHostNames.clear();
    availableHosts.clear();
    waitedHostQueue.clear();
    waitedHosts.clear();
  }

  /**
   * The time taken to download a chunk of pages from a host is multiplied
   * by this waitFactor to determine how soon a host can be accessed again.
   *
   * @param waitFactor
   */
  public void setWaitFactor(int waitFactor) {
    this.waitFactor = waitFactor;
  }

  public void setLastModifiedDb(LastModifiedDB lastModifiedDb) {
    this.lastModifiedDb = lastModifiedDb;
  }

  public int getCurrentSize() {
    return currentSize;
  }
}
