package org.supermind.crawl;

import java.util.LinkedList;
import java.util.TreeSet;
import java.util.zip.Checksum;
import java.util.zip.CRC32;
import java.net.URL;

/**
 * Queue of {@link ScheduledURL}s which belong to a particular host.
 */
public class HostQueue {

  /**
   * Protocol of this host. See {@link URL#getProtocol}.
   */
  private String protocol;

  /**
   * Name of this host. See {@link URL#getHost}.
   */
  private String name;

  /**
   * Pages in this host for crawling.
   */
  private LinkedList<ScheduledURL> pages = new LinkedList();

  /**
   * CRC32 checksums of URLs.
   */
  private TreeSet<Long> checksums = new TreeSet();

  /**
   * 
   */
  private Checksum crc = new CRC32();

  /**
   * Time (in milliseconds) when urls from this host can be continue to be fetched.
   * Used to implement a server politeness policy.
   */
  private long nextAvailable;


  /**
   * Ctor.
   * @param name Name of host
   * @param protocol Protocol
   */
  public HostQueue(String name, String protocol) {
    this.name = name;
    this.protocol = protocol;
  }

  public void add(ScheduledURL scheduledURL) {
    pages.add(scheduledURL);
    byte[] bytes = scheduledURL.url.toString().getBytes();
    crc.update(bytes, 0, bytes.length);
    checksums.add(crc.getValue());
    crc.reset();
  }

  /**
   * Pop {@link ScheduledURL}s for crawling. Number of ScheduledURLs popped is
   * {@link Fetcher#maxPagesPerConnection}.
   * @return ScheduledURL[] or null, if no more pages to fetch
   */
  public ScheduledURL[] pop() {
    int number = Math.min(pages.size(), Fetcher.maxPagesPerConnection);
    if(number < 1) return null;   // no more pages

    ScheduledURL[] array = new ScheduledURL[number];
//    pages.subList(0, number - 1).toArray(array);
    for (int i = 0; i < number; ++i) {
      array[i] = pages.removeFirst();
      byte[] bytes = array[i].url.toString().getBytes();
      crc.update(bytes, 0, bytes.length);
      checksums.remove(crc.getValue());
      crc.reset();
    }
    return array;
  }

  public boolean contains(URL url) {
    byte[] bytes = url.toString().getBytes();
    crc.update(bytes, 0, bytes.length);
    long value = crc.getValue();
    crc.reset();
    return checksums.contains(value);
  }

  public String getProtocol() {
    return protocol;
  }

  public String getName() {
    return name;
  }

  public LinkedList<ScheduledURL> getPages() {
    return pages;
  }

  public long getNextAvailable() {
    return nextAvailable;
  }

  public void setNextAvailable(long nextAvailable) {
    this.nextAvailable = nextAvailable;
  }

  public boolean equals(Object o) {
    if (this == o) return true;
    if (!(o instanceof HostQueue)) return false;

    final HostQueue hostQueue = (HostQueue) o;

    if (nextAvailable != hostQueue.nextAvailable) return false;
    if (!name.equals(hostQueue.name)) return false;
    if (protocol != null ? !protocol.equals(hostQueue.protocol) : hostQueue.protocol != null) return false;

    return true;
  }

  public int hashCode() {
    int result;
    result = (protocol != null ? protocol.hashCode() : 0);
    result = 29 * result + name.hashCode();
    result = 29 * result + (int) (nextAvailable ^ (nextAvailable >>> 32));
    return result;
  }


  public String toString() {
    return "HostQueue{" +
        "protocol='" + protocol + "'" +
        ", name='" + name + "'" +
        ", size='" + pages.size() + "'" +
        "}";
  }
}
