package org.supermind.crawl;

import org.apache.nutch.net.protocols.HttpDateFormat;
import org.apache.nutch.protocol.Content;
import org.apache.nutch.protocol.ProtocolOutput;
import org.supermind.crawl.util.LongLongPersister;

import java.io.IOException;
import java.net.URL;
import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Locale;
import java.util.logging.Level;
import java.util.zip.Checksum;

/**
 * Records URLs and their last modified times.
 */
public class LastModifiedFetchedURLs implements FetchedURLs, LastModifiedDB {
  private LongLongPersister persister;
  private Checksum checksum;

  public LastModifiedFetchedURLs() {
  }

  public void init() throws IOException {
    persister.init();
  }

  /**
   * Has this URL already been fetched?
   *
   * @param url
   * @return
   */
  public synchronized boolean contains(URL url) {
    long chksum = getChecksum(url);
    boolean contains = false;
    try {
      contains = persister.contains(chksum);
    } catch (IOException e) {
      e.printStackTrace();
    }
//    Fetcher.LOG.finer("Contains " + url + " : " + contains);
    return contains;
  }

  public void insert(ScheduledURL url, ProtocolOutput output) throws IOException {
    Content content = output.getContent();

    long time = System.currentTimeMillis(); // for urls with no last modified, use current time 
    String lastModified = content.getMetadata().getProperty("last-modified");
    if (lastModified != null) {                   // try parse last-modified
      time = getTime(lastModified, url.toString());           // use as time
    }

    long chksm = getChecksum(url.url);
    persister.add(chksm, time);
  }

  public ScheduledURL get(long id) {
    throw new UnsupportedOperationException("get() not supported");
  }
  
  public void close() throws IOException {
    persister.close();
  }

  public long getLastModified(URL url){
    try {
      long chksum = getChecksum(url);
      return persister.getValue(chksum);
    } catch (IOException e) {
      LOG.log(Level.WARNING, "Error encountered", e);
    }
    return -1;
  }

  public void setPersister(LongLongPersister persister) {
    this.persister = persister;
  }

  public void setChecksum(Checksum checksum) {
    this.checksum = checksum;
  }

  /**
   * Create a 64-bit checksum by merging a 32-bit host checksum with the url's
   * 32-bit checksum. By using host checksum as sig. bits, urls can be easily
   * sorted by host.
   *
   * @param url
   * @return
   */
  protected long getChecksum(URL url) {
    String host = url.getHost();
    byte[] bytes = host.getBytes();
    checksum.update(bytes, 0, bytes.length);
    long hostchecksum = checksum.getValue() << 32;
    checksum.reset();

    bytes = url.toString().getBytes();
    checksum.update(bytes, 0, bytes.length);
    long urlchecksum = checksum.getValue();
    checksum.reset();

    return hostchecksum | urlchecksum;
  }

  // Copied from Nutch's MoreIndexingFilter
  private static long getTime(String date, String url) {
    long time = -1;
    try {
      time = HttpDateFormat.toLong(date);
    } catch (ParseException e) {
      // try to parse it as date in alternative format
      String date2 = date;
      try {
        if (date.length() > 25) date2 = date.substring(0, 25);
        DateFormat df = new SimpleDateFormat("EEE, dd MMM yyyy HH:mm:ss", Locale.US);
        time = df.parse(date2).getTime();
      } catch (Exception e1) {
        try {
          if (date.length() > 24) date2 = date.substring(0, 24);
          DateFormat df = new SimpleDateFormat("EEE MMM dd HH:mm:ss yyyy", Locale.US);
          time = df.parse(date2).getTime();
        } catch (Exception e2) {
          LOG.warning(url + ": can't parse erroneous date: " + date);
        }
      }
    }
    return time;
  }

//  public static void main(String[] args) throws IOException {
//    LocalFileSystem nfs = new LocalFileSystem();
//    String filename = "c:/foo.txt";
//    SequenceFile.Writer writer = new SequenceFile.Writer(nfs, filename, LongWritable.class, NullWritable.class);
//    for (int i = 0; i < 2826; ++i) {
//      int value = (i * i);
//      if (i % 2 == 0) value -= i;
//      writer.append(new LongWritable(value), NullWritable.get());
//    }
//    writer.close();
//
//    SequenceFile.Sorter sorter = new SequenceFile.Sorter(nfs, new LongWritable.DecreasingComparator(), NullWritable.class);
//    String sortedname = "c:/foo.sorted.txt";
//    sorter.sort(filename, sortedname);
//
//    SequenceFile.Reader reader = new SequenceFile.Reader(nfs, sortedname);
//    MapFile.Writer mapwriter = new MapFile.Writer(nfs, "c:/temp/mapdir",
//        new LongWritable.DecreasingComparator(), NullWritable.class);
//    LongWritable w = new LongWritable();
//    while (reader.next(w)) {
//      System.out.println(w.get());
//      mapwriter.append(w, NullWritable.get());
//    }
//    reader.close();
//    mapwriter.close();
//  }

//  public static void main(String[] args) throws MalformedURLException {
//    URL url = new URL("http://www.networksecurityarchive.org/html/Security-Basics/2004-10/msg00189.html");
//    String host = url.getHost();
//    String s = url.toString();
//    int idx = s.indexOf(host) + host.length();
//
//    Checksum checksum = new CRC32();
//    byte[] bytes = s.getBytes();
//    checksum.update(bytes, 0, bytes.length);
//    long totalcheck = checksum.getValue();
//    System.out.println("URL:" + Long.toBinaryString(totalcheck));
//    checksum.reset();
//
//    bytes = host.getBytes();
//    checksum.update(bytes, 0, bytes.length);
//    long precheck = checksum.getValue();
//    System.out.println("Pre:" + Long.toBinaryString(precheck));
//    System.out.println("Pre:" + Long.toBinaryString(precheck << 32));
//    checksum.reset();
//
//    long result = (precheck << 32) | (totalcheck);
//    System.out.println("Tot:" + Long.toBinaryString(result));
//    System.out.println("Tot:" + Integer.toBinaryString((int) result));
//    System.out.println("Tot:" + Integer.toBinaryString((int) (result >>> 32)));
//  }
}
