package org.supermind.crawl.http;

import org.apache.nutch.protocol.Content;
import org.apache.nutch.protocol.ProtocolStatus;
import org.apache.nutch.util.GZIPUtils;
import org.apache.nutch.util.NutchConf;
import org.apache.nutch.util.mime.MimeType;
import org.apache.nutch.util.mime.MimeTypes;

import java.io.*;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.net.URL;
import java.util.*;
import java.util.logging.Level;

public class HttpResponse {
  static final String HTTP_TRANSFER_ENCODING = "Transfer-Encoding";
  static final String HTTP_CONTENT_ENCODING = "Content-Encoding";
  static final String HTTP_CONTENT_TYPE = "Content-Type";
  static final byte VERSION_1_0 = 0;
  static final byte VERSION_1_1 = 1;

  /**
   * A flag that tells if magic resolution must be performed
   */
  private final static boolean MAGIC =
      NutchConf.get().getBoolean("mime.type.magic", true);

  /**
   * Get the MimeTypes resolver instance.
   */
  private final static MimeTypes MIME =
      MimeTypes.get(NutchConf.get().get("mime.types.file"));

  /**
   * Map of host name to http version.
   */
  static Map<String, Byte> serverHttpVersion = new Hashtable();

  /**
   * Map of url to response code.
   */
  protected Map<String, Integer> codes = new HashMap();

  /**
   * Returns the response code.
   */
  public int getCode(String url) {
    Integer code = codes.get(url);
    if (code == null) {
      System.out.println("No code");
      Http.LOG.warning("No code found for " + url + " codemap:" + codes);
      return 500;
    }
    return code;
  }

  public Content get(String originalurl, final URL url) throws HttpException, IOException {
    Socket socket = null;
    URL u = url;
    Content content = null;

    try {
      if (!"http".equals(u.getProtocol()))
        throw new HttpException("Not an HTTP url:" + u);

//    if (Http.LOG.isLoggable(Level.FINE))
//      Http.LOG.fine("fetching " + url);

      // some servers will redirect a request with a host line like
      // "HostQueue: <hostname>:80" to "http://<hpstname>/<orig_path>"- they
      // don't want the :80...

      String host = u.getHost();
      int port = (u.getPort() == -1) ? 80 : u.getPort();
      socket = createSocket(host, port);

      OutputStream req = socket.getOutputStream();
      String path = getPath(u);
      byte[] reqBytes = buildRequest(u, path);

      req.write(reqBytes);
      req.flush();


      PushbackInputStream in = // process response
          new PushbackInputStream(new BufferedInputStream(socket.getInputStream(), Http.BUFFER_SIZE),
              Http.BUFFER_SIZE);

      String urlstr = u.toString();
      String orig = originalurl != null ? originalurl : urlstr;

      content = createContent(in, u, orig);
    } finally {
      if (socket != null) {
        socket.close();
        if (Http.LOG.isLoggable(Level.FINER)) Http.LOG.finer("Socket closed");
      }
    }

    return content;
  }

  /**
   * URLs MUST have the same host.
   *
   * @param origs
   * @param urls
   * @return
   * @throws IOException
   * @throws HttpException
   */
  public Content[] get(String[] origs, URL[] urls) {
    Socket socket = null;
    OutputStream req = null;
    Content[] result = new Content[urls.length];
    BitSet pipelinedUrls = new BitSet(urls.length);
    /**
     * Fetch the URLs that require Http 1.0 or for which we do not know the
     * server's Http version. All other URLs will be pipelined.
     */
    for (int i = 0; i < urls.length; ++i) {
      URL url = urls[i];
      if (!"http".equals(url.getProtocol())) {
        Http.LOG.warning("Not an HTTP url:" + url);
        result[i] = null;
      }

      if (Http.LOG.isLoggable(Level.FINE))
        Http.LOG.fine("fetching " + url);

      // some servers will redirect a request with a host line like
      // "HostQueue: <hostname>:80" to "http://<hpstname>/<orig_path>"- they
      // don't want the :80...

      String host = url.getHost();

      int port;
      if (url.getPort() == -1) {
        port = 80;
      } else {
        port = url.getPort();
      }

      Byte httpVer = serverHttpVersion.get(host);
      String urlstr = url.toString();
      String orig = origs != null ? origs[i] : urlstr;
      /**
       * Do not pipeline when server has Http 1.0.
       * Since RobotsRuleParser will always attempt to fetch robots.txt _before_
       * we get here, the server's HTTP version should always be known
       */
      if (httpVer == null                           // don't know the servers HTTP version for some reason
          || httpVer.byteValue() == VERSION_1_0) {  // or HTTP 1.0 which doesn't officially support pipelining
        if (Http.LOG.isLoggable(Level.FINER))
          Http.LOG.finer("Not pipelining " + urlstr + " HTTP version:" + httpVer);
        try {
          result[i] = get(orig, url);
        } catch (Exception e) {
          Http.LOG.log(Level.WARNING, "Error encountered getting " + url, e);
          result[i] = null;
        }

      } else {    // pipeline: requests are written continuously to the socket
        try {
          if (socket == null) socket = createSocket(host, port);
          req = socket.getOutputStream(); // make request

          String path = getPath(url);
          byte[] reqBytes = buildRequest(url, path);
          req.write(reqBytes);
          req.flush();

          pipelinedUrls.set(i);
        } catch (IOException e) {
          Http.LOG.log(Level.WARNING, "IOException pipelining req:" + url, e);
          result[i] = null;
        }
      }
    }

    // Process pipelined requests
    PushbackInputStream in = null;
    try {
      in = new PushbackInputStream(new BufferedInputStream(socket.getInputStream(), Http.BUFFER_SIZE), Http.BUFFER_SIZE);
    } catch (IOException e) {
      Http.LOG.log(Level.WARNING, "IOException getting socket inputstream", e);
    }

    if (in != null) {
      for (int i = 0; i < urls.length; ++i) {
        if (!pipelinedUrls.get(i)) continue;

        URL url = urls[i];
        String base = url.toString();
        String orig = origs != null ? origs[i] : base;
        try {
          result[i] = createContent(in, url, orig);
        } catch (Exception e) { // TODO: handle SocketTimeoutException separately?

          // pipelined URLs are prone to error. Let's see if we can fetch manually
          try {
            result[i] = get(orig, url);
          } catch (Exception e2) {   // we failed, but not for lack of trying
            Http.LOG.log(Level.WARNING, "Error encountered getting " + url, e2);
            result[i] = null;
          }
        }
      }       // end for
    }

    return result;
  }

  private Content createContent(PushbackInputStream in, URL url, String orig)
      throws IOException, HttpException {
    Properties headers = new Properties();
    String base = url.toString();
    byte[] contentBytes = getResponse(in, url, headers);
    String contentType = headers.getProperty(HTTP_CONTENT_TYPE);
    if (contentType == null) {
      MimeType type = null;
      if (MAGIC) {
        type = MIME.getMimeType(orig, contentBytes);
      } else {
        type = MIME.getMimeType(orig);
      }
      if (type != null) {
        contentType = type.getName();
      } else {
        contentType = "";
      }
    }
    Content content = new Content(base, orig, contentBytes, contentType, headers);
    redirectIfNecessary(url, content);
    return content;
  }

  /**
   * Redirects are not pipelined, nor are existing sockets re-used (since there
   * is a decent chance that the redirect will be to another host).
   *
   * @param url
   * @param c
   * @return
   * @throws IOException
   * @throws HttpException
   */
  private Content redirectIfNecessary(URL url, final Content c) throws IOException, HttpException {
    Content content = c;
    int redirects = 0;
    Integer status = codes.get(url.toString());
    while (true) {
      Properties headers = content.getMetadata();
      String location = headers.getProperty("Location");
// some broken servers, such as MS IIS, use lowercase header name...
      if (location == null) location = headers.getProperty("location");

      if (status == null || status.intValue() < 300 // redirect
          || status.intValue() >= 400 || location == null) {
        break;
      } else if (redirects == Http.MAX_REDIRECTS) {
        codes.put(url.toString(), ProtocolStatus.REDIR_EXCEEDED);
        break;
      }

      redirects++;
      URL redirecturl = new URL(url, location);
      if (Http.LOG.isLoggable(Level.FINE)) Http.LOG.fine("Redirecting " + url.toString() + " to " + redirecturl.toString());
      int port = (redirecturl.getPort() == -1) ? 80 : redirecturl.getPort();
      Socket redirectsocket = null;
      try {
        redirectsocket = createSocket(redirecturl.getHost(), port);
        OutputStream redout = redirectsocket.getOutputStream();
        PushbackInputStream redin = // process response
            new PushbackInputStream(new BufferedInputStream(redirectsocket.getInputStream(), Http.BUFFER_SIZE),
                Http.BUFFER_SIZE);

        String path = getPath(redirecturl);
        byte[] reqBytes = buildRequest(redirecturl, path);
        redout.write(reqBytes);
        redout.flush();

// set new content params
        headers.clear();

//        content.setUrl(redirecturl.toString()); //FIXME: content has no setter for url
        byte[] contentBytes = getResponse(redin, redirecturl, headers);
        String contentType = headers.getProperty(HTTP_CONTENT_TYPE);
        if (contentType == null) {
          MimeType type = null;
          if (MAGIC) {
            type = MIME.getMimeType(content.getBaseUrl(), contentBytes);
          } else {
            type = MIME.getMimeType(content.getBaseUrl());
          }
          if (type != null) {
            contentType = type.getName();
          } else {
            contentType = "";
          }

// FIXME: remove when setUrl is added
          content = new Content(redirecturl.toString(), content.getBaseUrl(),
              contentBytes, contentType, headers);
        }
      } finally {
        if (redirectsocket != null) redirectsocket.close();
      }
      status = codes.get(redirecturl.toString());
    }     // end while

    return content;
  }

  private Socket renewSocket(Socket socket, String host, int port) throws IOException, HttpException {
    if (Http.LOG.isLoggable(Level.FINER)) Http.LOG.finer("Renewing socket to " + host + ":" + port);
    if (socket != null) {
      socket.close();
      if (Http.LOG.isLoggable(Level.FINER)) Http.LOG.finer("Socket closed");
    }
    socket = createSocket(host, port);
    return socket;
  }

  private Socket renewSocket(Socket socket, int i, Content[] result, URL[] urls, String[] origs, String host, int port) throws IOException, HttpException {
    if (Http.LOG.isLoggable(Level.FINER)) Http.LOG.finer("Renewing socket to " + host + ":" + port);
    if (socket != null) {
      PushbackInputStream in = // process response
          new PushbackInputStream(new BufferedInputStream(socket.getInputStream(), Http.BUFFER_SIZE),
              Http.BUFFER_SIZE);
      for (int j = i - 1; j >= 0; j--) {
        if (result[j] != null) continue;

        URL u = urls[j];
        String base = u.toString();
        String o = origs != null ? origs[j] : base;

        result[j] = createContent(in, u, o);
        if (Http.LOG.isLoggable(Level.FINER)) Http.LOG.finer(u.toString() + " flushed");
      }
      socket.close();
      if (Http.LOG.isLoggable(Level.FINER)) Http.LOG.finer("Socket closed");
    }
    socket = createSocket(host, port);
    return socket;
  }


  private Socket createSocket(String host, int port) throws IOException {
    if (Http.LOG.isLoggable(Level.FINER)) Http.LOG.finer("Creating socket to " + host + ":" + port);
    Socket socket;
    socket = new Socket();                    // create the socket
    socket.setSoTimeout(Http.TIMEOUT);

// connect
    String sockHost = Http.PROXY ? Http.PROXY_HOST : host;
    int sockPort = Http.PROXY ? Http.PROXY_PORT : port;
    InetSocketAddress sockAddr = new InetSocketAddress(sockHost, sockPort);
    socket.connect(sockAddr, Http.TIMEOUT);
    return socket;
  }

  private String getPath(URL url) {
    String path = "".equals(url.getFile()) ? "/" : url.getFile();
    return path;
  }

  private byte[] buildRequest(URL url, String path) {
    String host = url.getHost();
    int port;
    String portString;
    if (url.getPort() == -1) {
      port = 80;
      portString = "";
    } else {
      port = url.getPort();
      portString = ":" + port;
    }
    StringBuffer reqStr = new StringBuffer("GET ");
    if (Http.PROXY) {
      reqStr.append(url.getProtocol() + "://" + host + portString + path);
    } else {
      reqStr.append(path);
    }

    reqStr.append(" HTTP/1.1\r\n");

    reqStr.append("Host: ");
    reqStr.append(host);
    reqStr.append(portString);
    reqStr.append("\r\n");

//    reqStr.append("If-Modified-Since: ");
//    reqStr.append(lastModified);
//    reqStr.append("\r\n");

    reqStr.append("Accept-Encoding: x-gzip, gzip\r\n");

    if ((Http.AGENT_STRING == null) || (Http.AGENT_STRING.length() == 0)) {
      Http.LOG.severe("User-agent is not set!");
    } else {
      reqStr.append("User-Agent: ");
      reqStr.append(Http.AGENT_STRING);
      reqStr.append("\r\n");
    }
//    Byte serverVer = serverHttpVersion.get(host);
//    if(serverVer != null && serverVer == VERSION_1_0)
//      reqStr.append("Connection: Keep-Alive\r\n");
    reqStr.append("\r\n");
    byte[] reqBytes = reqStr.toString().getBytes();

    return reqBytes;
  }

  private byte[] getResponse(PushbackInputStream in, URL url, Properties headers) throws IOException, HttpException {

    StringBuffer statusLine = new StringBuffer();

    boolean haveSeenNonContinueStatus = false;
    while (!haveSeenNonContinueStatus) {
// parse status code line
      int code = parseStatusLine(in, statusLine, url);
      codes.put(url.toString(), code);
// parse headers
      headers.putAll(parseHeaders(in, statusLine));
      haveSeenNonContinueStatus = code != 100; // 100 is "Continue"
    }

    if (Http.LOG.isLoggable(Level.FINEST)) Http.LOG.finest("Headers:" + headers);

    String te = headers.getProperty(HTTP_TRANSFER_ENCODING);
    byte[] content = te != null && te.equals("chunked")
        ? readChunkedContent(in, new StringBuffer()) :
        readPlainContent(in, headers);

    String contentEncoding = headers.getProperty(HTTP_CONTENT_ENCODING);
    if ("gzip".equals(contentEncoding) || "x-gzip".equals(contentEncoding)) {
      Http.LOG.finest("uncompressing....");
      byte[] compressed = content;

      content = GZIPUtils.unzipBestEffort(compressed, Http.MAX_CONTENT);

      if (content == null)
        throw new HttpException("unzipBestEffort returned null");

      if (Http.LOG.isLoggable(Level.FINEST))
        Http.LOG.finest("fetched " + compressed.length
            + " bytes of compressed content (expanded to "
            + content.length + " bytes) from " + url);
    } else {
      if (Http.LOG.isLoggable(Level.FINEST))
        Http.LOG.finest("fetched " + content.length + " bytes from " + url);
    }
//    System.out.println(new String(content));
    return content;
  }

  private byte[] readPlainContent(InputStream in, Properties headers)
      throws HttpException, IOException {

    int contentLength = Integer.MAX_VALUE;    // get content length
    String contentLengthString = headers.getProperty("Content-Length");
    if (contentLengthString != null) {
      contentLengthString = contentLengthString.trim();
      try {
        contentLength = Integer.parseInt(contentLengthString);
      } catch (NumberFormatException e) {
        throw new HttpException("bad content length: " + contentLengthString);
      }
    }
    if (Http.MAX_CONTENT >= 0
        && contentLength > Http.MAX_CONTENT)   // limit download size
      contentLength = Http.MAX_CONTENT;

    ByteArrayOutputStream out = new ByteArrayOutputStream(Http.BUFFER_SIZE);
    byte[] bytes = new byte[Http.BUFFER_SIZE];
    int length = 0;                           // read content
    for (int i = in.read(bytes); i != -1; i = in.read(bytes)) {

      out.write(bytes, 0, i);
      length += i;
      if (length >= contentLength)
        break;
    }
    return out.toByteArray();
  }

  private byte[] readChunkedContent(PushbackInputStream in,
                                    StringBuffer line)
      throws HttpException, IOException {
    boolean doneChunks = false;
    int contentBytesRead = 0;
    byte[] bytes = new byte[Http.BUFFER_SIZE];
    ByteArrayOutputStream out = new ByteArrayOutputStream(Http.BUFFER_SIZE);

    while (!doneChunks) {
//      if (Http.LOG.isLoggable(Level.FINER)) Http.LOG.finer("Http: starting chunk");

      readLine(in, line, false);

      String chunkLenStr;
// LOG.fine("chunk-header: '" + line + "'");

      int pos = line.indexOf(";");
      if (pos < 0) {
        chunkLenStr = line.toString();
      } else {
        chunkLenStr = line.substring(0, pos);
// LOG.fine("got chunk-ext: " + line.substring(pos+1));
      }
      chunkLenStr = chunkLenStr.trim();
//      System.out.println("Chunk len:" + chunkLenStr);
      int chunkLen;
      try {
        chunkLen = Integer.parseInt(chunkLenStr, 16);
      } catch (NumberFormatException e) {
        throw new HttpException("bad chunk length: " + line.toString());
      }

      if (chunkLen == 0) {
        doneChunks = true;
        break;
      }

      if ((contentBytesRead + chunkLen) > Http.MAX_CONTENT)
        chunkLen = Http.MAX_CONTENT - contentBytesRead;

// read one chunk
      int chunkBytesRead = 0;
      while (chunkBytesRead < chunkLen) {

        int toRead = (chunkLen - chunkBytesRead) < Http.BUFFER_SIZE ?
            (chunkLen - chunkBytesRead) : Http.BUFFER_SIZE;
        int len = in.read(bytes, 0, toRead);

        if (len == -1)
          throw new HttpException("chunk eof after " + contentBytesRead
              + " bytes in successful chunks"
              + " and " + chunkBytesRead
              + " in current chunk");

// DANGER!!! Will printed GZIPed stuff right to your
// terminal!
// LOG.fine("read: " +  new String(bytes, 0, len));

        out.write(bytes, 0, len);
        chunkBytesRead += len;
      }

      readLine(in, line, false);

    }

    if (!doneChunks) {
      if (contentBytesRead != Http.MAX_CONTENT)
        throw new HttpException("chunk eof: !doneChunk && didn't max out");
      return null;
    }

    parseHeaders(in, line);
    return out.toByteArray();

  }

  private int parseStatusLine(PushbackInputStream in, StringBuffer line, URL url)
      throws IOException, HttpException {
    readLine(in, line, false);

    if (line.length() == 0) {
      Http.LOG.warning("Empty status line:" + url);
      readLine(in, line, false);
    }

    char httpVer = line.charAt(7);
    byte serverVersion = httpVer == '1' ? VERSION_1_1 : VERSION_1_0;
    serverHttpVersion.put(url.getHost(), serverVersion);

    int codeStart = line.indexOf(" ");
    int codeEnd = line.indexOf(" ", codeStart + 1);

// handle lines with no plaintext result code, ie:
// "HTTP/1.1 200" vs "HTTP/1.1 200 OK"
    if (codeEnd == -1)
      codeEnd = line.length();

    int code;
    try {
      code = Integer.parseInt(line.substring(codeStart + 1, codeEnd));
    } catch (NumberFormatException e) {
      throw new HttpException("bad status line '" + line
          + "': " + e.getMessage(), e);
    }

    return code;
  }


  private void processHeaderLine(StringBuffer line, TreeMap headers)
      throws HttpException {
    int colonIndex = line.indexOf(":");       // key is up to colon
    if (colonIndex == -1) {
      int i;
      for (i = 0; i < line.length(); i++)
        if (!Character.isWhitespace(line.charAt(i)))
          break;
      if (i == line.length())
        return;
      throw new HttpException("No colon in header:" + line);
    }
    String key = line.substring(0, colonIndex);

    int valueStart = colonIndex + 1;            // skip whitespace
    while (valueStart < line.length()) {
      int c = line.charAt(valueStart);
      if (c != ' ' && c != '\t')
        break;
      valueStart++;
    }
    String value = line.substring(valueStart);

    headers.put(key, value);
  }

  private Map parseHeaders(PushbackInputStream in, StringBuffer line)
      throws IOException, HttpException {
    TreeMap headers = new TreeMap(String.CASE_INSENSITIVE_ORDER);
    return parseHeaders(in, line, headers);
  }

// Adds headers to an existing TreeMap
  private Map parseHeaders(PushbackInputStream in, StringBuffer line,
                           TreeMap headers)
      throws IOException, HttpException {
    while (readLine(in, line, true) != 0) {

      // handle HTTP responses with missing blank line after headers
      int pos;
      if (((pos = line.indexOf("<!DOCTYPE")) != -1)
          || ((pos = line.indexOf("<HTML")) != -1)
          || ((pos = line.indexOf("<html")) != -1)) {

        in.unread(line.substring(pos).getBytes("UTF-8"));
        line.setLength(pos);

        try {
          processHeaderLine(line, headers);
        } catch (Exception e) {
// fixme:
          e.printStackTrace();
        }

        return headers;
      }

      processHeaderLine(line, headers);
    }
    return headers;
  }

  private static int readLine(PushbackInputStream in, StringBuffer line,
                              boolean allowContinuedLine)
      throws IOException {
    line.setLength(0);
    for (int c = in.read(); c != -1; c = in.read()) {
      switch (c) {
        case '\r':
          if (peek(in) == '\n') {
            in.read();
          }
        case '\n':
          if (line.length() > 0) {
// at EOL -- check for continued line if the current
// (possibly continued) line wasn't blank
            if (allowContinuedLine)
              switch (peek(in)) {
                case ' ':
                case '\t':                   // line is continued
                  in.read();
                  continue;
              }
          }
          return line.length();      // else complete
        default :
          line.append((char) c);
      }
    }
    throw new EOFException();
  }

  private static int peek(PushbackInputStream in) throws IOException {
    int value = in.read();
    in.unread(value);
    return value;
  }
}
